﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
        GO_AHEAD, //IA006 - André A. Silva
		MOVE_IT,//IA006 - André A. Silva
		SACK_IT, //IA006 - André A. Silva
		EAT_IT //IA006 - André A. Silva
    }

    public class ClarionAgent
    {
        #region Constants

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";


		/// <summary>
		/// IA006 - André A. Silva - Constante que representa que existe uma jóia próxima a criatura
		/// </summary>
		private String DIMENSION_JEWEL_AHEAD = "JewelAhead";
		/// <summary>
		/// IA006 - André A. Silva - Constante que representa que existe comida próxima a criatura
		/// </summary>
		private String DIMENSION_FOOD_AHEAD = "FoodAhead";




        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;


		/// <summary>
		/// IA006 - André A. Silva - Perception input para indicar uma jóia no campo de visão
		/// </summary>
		private DimensionValuePair InputJewelAhead;
		/// <summary>
		/// IA006 - André A. Silva - Perception input para indicar comida no campo de visão
		/// </summary>
		private DimensionValuePair InputFoodAhead;





        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        private ExternalActionChunk OutputGoAhead;


		/// <summary>
		/// IA006 - André A. Silva - Output action que faz o agente mover-se e alimentar-se
		/// </summary>
		private ExternalActionChunk OutputFoodAhead;
		/// <summary>
		/// IA006 - André A. Silva - Output action que faz o agente mover-se e ensacar um objeto
		/// </summary>
		private ExternalActionChunk OutputJewelAhead;



        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
			InputJewelAhead = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_AHEAD); //IA006 - André A. Silva
			InputFoodAhead = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_FOOD_AHEAD);//IA006 - André A. Silva


            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
			OutputJewelAhead = World.NewExternalActionChunk(ClarionAgentActionType.SACK_IT.ToString());//IA006 - André A. Silva
			OutputFoodAhead = World.NewExternalActionChunk(ClarionAgentActionType.EAT_IT.ToString());//IA006 - André A. Silva
			OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());//IA006 - André A. Silva


            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
		{
			// Create Colission Wall Rule
			SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
			FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit (ruleAvoidCollisionWall);

// Create Colission To Go Ahead
            SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);
// Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoAhead);

			//IA006 - André A. Silva - regra para mover-se e ensacar uma jóia;
			SupportCalculator JewelAheadSupportCalculator = FixedRuleDelegateToJewelAhead;
			FixedRule ruleJewelAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputJewelAhead, JewelAheadSupportCalculator);
			CurrentAgent.Commit(ruleJewelAhead);

			/// IA006 - André A. Silva - regra para mover-se e alimentar-se
			SupportCalculator FoodAheadSupportCalculator = FixedRuleDelegateToFoodAhead;
			FixedRule ruleFoodAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputFoodAhead, FoodAheadSupportCalculator);
			CurrentAgent.Commit(ruleFoodAhead);



            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

			Creature criatura = (Creature) sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_CREATURE)).First();
			//creature.PrintLeaflet();


			//IA006 - André A. Silva - Detecta se temos jóias no campo de visão
			Boolean jewelAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_JEWEL && criatura.inLeaflet(item.Material.Color))).Any();
			//IA006 - André A. Silva - Detecta se temos comida no campo de visão
			Boolean foodAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_FOOD || item.CategoryId == Thing.categoryPFOOD || item.CategoryId == Thing.CATEGORY_NPFOOD )).Any();
			// Detect if we have a wall ahead - OU nada próximo foi captado no campo de visão (IA006)
			Boolean wallAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 61)).Any();


			//IA006 - André A. Silva - determinação da regra e solução de conflitos.
			if (!jewelAhead && !foodAhead) {
//				Console.WriteLine ("1- JewelAhead = " + jewelAhead.ToString ());
//				Console.WriteLine ("1- FoodAhead = " + foodAhead.ToString ());
//				Console.WriteLine ("1- WallAhead = " + wallAhead.ToString ());

				si.Add (InputJewelAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
			} else {
				if (jewelAhead && foodAhead) {
//					Console.WriteLine ("2- JewelAhead = " + jewelAhead.ToString ());
//					Console.WriteLine ("2- FoodAhead = " + foodAhead.ToString ());
//					Console.WriteLine ("2- WallAhead = " + wallAhead.ToString ());

					si.Add (InputJewelAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
					si.Add (InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add (InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
				} else {
					if (jewelAhead) {
//						Console.WriteLine ("3- JewelAhead = " + jewelAhead.ToString ());
//						Console.WriteLine ("3- FoodAhead = " + foodAhead.ToString ());
//						Console.WriteLine ("3- WallAhead = " + wallAhead.ToString ());

						si.Add (InputJewelAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
						si.Add (InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
						si.Add (InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
					} else {
						if (foodAhead) {
//							Console.WriteLine ("4- JewelAhead = " + jewelAhead.ToString ());
//							Console.WriteLine ("4- FoodAhead = " + foodAhead.ToString ());
//							Console.WriteLine ("4- WallAhead = " + wallAhead.ToString ());

							si.Add (InputJewelAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
							si.Add (InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
							si.Add (InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
						}
					}
				}
			}


            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

		private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
		{
			// See partial match threshold to verify what are the rules available for action selection
			return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target)
		{
			// Here we will make the logic to go ahead
			return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
		}


		//IA006 - André A. Silva - mover-se para a jóia e pegá-la
		private double FixedRuleDelegateToJewelAhead(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputJewelAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}


		//IA006 - André A. Silva - mover-se para a comida e alimentar-se
		private double FixedRuleDelegateToFoodAhead(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}




        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
